#include "GenomeMixer.h"
#include "CellData.h"
#include "MouseCell.h"
#include "MouseIO.h"
#include "PurebredData.h"

#include <qstring.h>

#include <vector>
#include <algorithm>
#include <math.h>

using namespace std;


// ================================================================
// GenomeMixer    (c) 2003, Alex Williams
// All rights reserved.
// Send email correspondence to Robert W. Williams <rwilliam@nb.utmem.edu>
// at the University of Tennesse Health Science Center

// This project was funded jointly by the National Institute of Mental Health, National Institute on Drug Abuse,
// and the National Science Foundation (award P20-MH 62009 to KFM and RW)

// This source code may be used according to the terms of the GNU General Public License (GPL).
// See www.gnu.org for specifics. Essentially, the source may be freely used as long as the source code
// for whatever program the source is also made freely available. If you want to use this source
// in a closed-source program, you should email us for licensing options.

// Disclaimer: This code comes with no warranty at all, and there is no guarantee that it will
// work as claimed in the documentation. Hopefully, it will be useful anyway.

// Note that GenomeMixer makes use of the Trolltech Qt toolkit for its window-managing needs.
// There is very little use of Qt data structures (aside from QString) in the actual processing code,
// but **you will need to have Qt installed in order to compile this code.** Check <www.trolltech.com>
// for details; there should be a free GPL-software-only version of Qt available on the Trolltech
// web site.

// Any redistributions of this software in code or binary form must include the
// above copyright notice, licensing information, disclaimer, and this notice itself.
// Basically, all the comments at the top of this file must be retained.
// ================================================================

/*
  Some information about this program:  
  
  GenomeMixer was originally a Perl program.
  
  GenomeMixer.cpp used to be the main part of the Perl program. Now, it has become an object that handles nearly all of the genome-mixing settings.
  Each GenomeMixer object also has a pointer to a class called "MarkerCollection," which is, unsurprisingly, a class that is a collection of markers.
  
  MouseIO is a class, but really it is just a collection of functions, since all the methods/functions in MouseIO are "static" (object-wide; that is, there is nothing to be gained in instantiating a MouseIO object. Calls can be made with MouseIO::doThisFunction(param1, param2) ).
  
  Chromsomes are each a "Chromosome" object, which takes care of dealing with all the genetic material in a Chromosome. There are also several "Chromosome" class variables, keeping track of things like total number of chromosomes, and the indices for the X and Y chromosomes. If this program was ever written to support more than one GenomeMixer open at one time, these class variables would have to be moved to a "ChromosomeCollection" or perhaps put in the GenomeMixer object.
  
  A Chromosome is comprised of many GeneRanges. Each GeneRange object has a start, end, and a value. The start and end are the base-pair positions, and the value is the genotype for that "range" of the genes. So the sequence AAAAABBBBB would be represented as two GeneRanges, one with start=0, end=4 (not 5), and value=A, and the other would be start=5, end=9 (not 10), and value=B. In reality, value=A would probably be "value=0" or something, since the value is a long integer, not a character. Sometimes, the sequence AAAAAAAAAAA can be represented by multiple GeneRanges, in case there is recombination (say, AAAA to AABB, and then later back to AAAA). We don't consolodate GeneRanges, because we want the option to keep track of where all the breakpoints occurred, even if they subseqently disappear. This is not done for now, however.
  
  PurebredData contains all the names & symbols for the purebred parents. All of this data used to be in GenomeMixer, but now a GenomeMixer has a pointer to its purebred data.
  
  */





GenomeMixer::GenomeMixer() {
	// The constructor for a GenomeMixer object
	
	clearCellDataTable();
	Chromosome::setNumberOfChromosomes(0);
	
	defaultLitterSize = 0;
	defaultLitterStDev = 0;
	myNumGenerations = 0;
	myNumCages = 0;
	
	noRecombDistBases = 20000000; // just a default...
	interferenceModel = 1; // default: abs inteference
	recombModel = 1; // default: linear recombination
	
	markerColPtr = NULL;
	mice = new MouseCell("Global Cell");
}


void GenomeMixer::setTableSize(int cages, int generations) {
	// Sizes the breedingTable and miceTable to the proper values, so they can store all
	// the data that we will want.
	assert(cages >= 0 && generations >= 0);
	assert(cages <= kTABLE_SIZE && generations <= kTABLE_SIZE);
	
	myNumCages       = cages;
	myNumGenerations = generations;
	
	//cout << "Number of cages is " << myNumCages << " and num generations is " << myNumGenerations << endl;
	
	//myBreedingTable = new string[cages][generations];
	//myMiceTable     = new MouseCell[cages][generations];
}


void GenomeMixer::setBreedingTableElem(int cage, int gen, string x) {
	// The breeding table is all in text, nothing fancy and data-structure-y
	myBreedingTable[cage][gen] = x;
	//cout << "x was: " << x << ", and the table is: " << myBreedingTable[cage][gen] << ".\n";
}
string GenomeMixer::getBreedingTable(int cage, int gen) {
	assert(cage >= 0 && gen >= 0 && cage < numCages() && gen < numGenerations());
	return myBreedingTable[cage][gen];
}



MouseCell* GenomeMixer::getMiceTable(int cage, int gen) {
	assert(cage >= 0 && gen >= 0 && cage < numCages() && gen < numGenerations());
	return &(myMiceTable[cage][gen]);
}



void GenomeMixer::setCellData(int cage, int gen, CellData *theData) {  
	assert(cage >= 0 && gen >= 0 && cage < numCages() && gen < numGenerations());  
	
	// If there is already something there, delete it
	if (myCellDataTable[cage][gen]) {
		delete (myCellDataTable[cage][gen]);
		myCellDataTable[cage][gen] = NULL; // <-- and set to null!
	}
	
	// Now, fill in the cell, if there is anything to fill it in with
	if (theData != NULL) {
		myCellDataTable[cage][gen] = new CellData(theData); // <-- a table of pointers in the first place
	}
}

CellData* GenomeMixer::getCellData(int cage, int gen) {
	assert(cage >= 0 && gen >= 0 && cage < numCages() && gen < numGenerations());
	return myCellDataTable[cage][gen]; // <-- returns a pointer
}

void GenomeMixer::clearCellDataTable() {
	// Completely clears the cell data table, and frees memory as well
	for (int i = 0; i < kTABLE_SIZE; i++) {
		for (int k = 0; k < kTABLE_SIZE; k++) {
			if (myCellDataTable[i][k]) {
				delete myCellDataTable[i][k];	
			}
			myCellDataTable[i][k] = NULL;
		}
	}    
}

// Functions having to do with purebreds/parental strains

PurebredData* GenomeMixer::purePtr(void) {
	return &pure;
}

void   GenomeMixer::clearAllParentalStrains(void) {
	pure.clearAll();
	// Erases all the ancestral strain (purebred) information
}

void   GenomeMixer::addParentalStrain(string strainName, string symbol) { 
	pure.addPurebred(strainName, symbol);// Adds a (purebred) parental strain
}
int    GenomeMixer::numParentalStrains() {
	return pure.numPurebreds();
}
void GenomeMixer::makePurebreds(void) {
	pure.makePurebredMice();
}

string GenomeMixer::getPurebredSymbol(int index) {
	try {
		return *(pure.getPureSymbolPtr(int(index)));
	} catch (const string *e) {
		cout << "ERROR AT " << e;
		throw e;
	}
}
string GenomeMixer::getPurebredName(int index) {
	try {
		return *(pure.getPureNamePtr(int(index)));
	} catch (const string *e) {
		cout << "ERROR AT " << e;
		throw e;
	}
}    
Mouse* GenomeMixer::getPurebredPtr(int index, bool sex) {
	// Returns a pointer to a purebred parental strain mouse
	return pure.getPurePtr(index, sex);
}
// End of purebred strain functions




void GenomeMixer::runSimulation(void) {
	try {
		clearExportList(); // First, clear the current list of export mice...
		// Run breeding simulation (go through the breedingTable)
		for (int gen = 0; gen < numGenerations(); gen++) {
			for (int cage = 0; cage < numCages(); cage++) {
				
				// Note that "decodeCell" still expects a string, so that's what it gets. We can change this later.
				MouseCellIngredients cellIngred;
				if ((getCellData(cage, gen) != NULL) &&
					MouseIO::decodeCell(&cellIngred, getCellData(cage, gen)->toString(&(this->pure)), this)) { // "this" is a pointer to the current GenomeMixer object
					//decodeCell returns a true value if it has information for us to make a litter with, false otherwise
					myMiceTable[cage][gen] = *GenomeMixer::NewLitter(&cellIngred);
				}    
			}
		}    
	} catch (const string *e) {
		throw e;
	}
}

int    GenomeMixer::numCages() {
	return myNumCages; // <-- public accessor for private var
} 

int    GenomeMixer::numGenerations() {
	return myNumGenerations;  // <-- public accessor for private var
}

void   GenomeMixer::clearExportList(void) {
	miceToExport.clear();
	miceToExport.reserve(0);
}

Mouse* GenomeMixer::mouseToExport(int index) {
	assert(index >= 0 && index < long(miceToExport.size()));
	return miceToExport[index]; // <-- public accessor for private vector
}

int    GenomeMixer::numMiceToExport(void) {
	return miceToExport.size(); // <-- public accessor
}

// ******** ******** ******** ******** ******** ******** ******** ******** ******** ******** ********
//							Make One Mouse
// ******** ******** ******** ******** ******** ******** ******** ******** ******** ******** ********
Mouse* GenomeMixer::MakeOneMouse(char gender, Mouse *mother, Mouse *father, int litterNumber, int siblingNumber) {
	try {
		// litterNumber: which litter is this mouse in? (first, second, etc...)
		// siblingNumber: from 1 to the size of the litter
		
		// gender is 'u' for unassigned (means, randomly assign a sex), 'm' for male, and 'f' for female
		if (!(gender == 'u' || gender == 'f' || gender == 'm')) {
			cerr << "Error: mouse was called with gender other than 'u', 'f', or 'm'." << endl;
			throw new string("Programming Error: A mouse was called with\na gender other than 'u', 'f', or 'm', in GenomeMixer::MakeOneMouse.");
		}
		
		if (mother == NULL || father == NULL) {
			cerr << "Error: mouse was called with a NULL parent." << endl;
			throw new string("A mouse was called with\na NULL parent, in GenomeMixer::MakeOneMouse.\nThis could mean that a cell requested a mouse that hasn't been\ngenerated yet (for example, reequesting that a\ngeneration 5 mouse be a parent for generation 2 mouse),\nor a cell could have requested a parent from a cell in which no mice of that sex\nwere available (for example, requesting a parent\nfrom a cell that generated 0 mice total).");
		}	
		
		assert(litterNumber >= -1); // <-- NOTE: right now, litter number isn't used, however
		
		Mouse *newMouse = new Mouse("New", kUNDEFINED, kUNDEFINED, mother, father, gender);
		
		//char motherIDstr[32];	sprintf(motherIDstr, "%ld", mother->ID());
		//char fatherIDstr[32];	sprintf(fatherIDstr, "%ld", father->ID());
		//char siblingNumStr[32];	sprintf(siblingNumStr, "%d", (siblingNumber+1));
		char sexChar = (newMouse->isMale() ? 'm' : 'f');
		
		string nameStr = string("(") + QString::number(mother->ID()) + string(" x ") + QString::number(father->ID()) + string(": ") + QString::number(siblingNumber+1) + string(" ") + sexChar + string(")");
		
		newMouse->setName(nameStr);
		
		// +-- give this mouse some genetic information...
		
		GenomeMixer::ChromosomeCombine(newMouse);
		
		// +-- check to see if mouse has any fatal traits
		// |
		// +--
		
		// random chance of mouse dying, offset by advantageous traits?
		// |
		// +--
		
		return newMouse;
	} catch (const string *e) {
		throw e;
	}
}

// ******** ******** ******** ******** ******** ******** ******** ******** ******** ******** ********
//							Mouse Blender (whirr)
// ******** ******** ******** ******** ******** ******** ******** ******** ******** ******** ********
MouseCell* GenomeMixer::NewLitter(MouseCellIngredients *in) {
	try {
		MouseCell *newLitter = new MouseCell(in->mother->name() + in->father->name());
		int       miceMadeSoFar = 0;
		
		// Make the specified number of female mice (if any)
		for (int i = 0;  i < in->numFemales;   i++, miceMadeSoFar++) {
			Mouse *newMouse = MakeOneMouse('f', in->mother, in->father, -1, miceMadeSoFar);
			if (in->shouldBeExported)  addExport(newMouse);
			newLitter->addMouse(newMouse);
		}
		
		// Make the specified number of male mice (if any)
		for (int i = 0;  i < in->numMales;   i++, miceMadeSoFar++) {
			Mouse *newMouse = MakeOneMouse('m', in->mother, in->father, -1, miceMadeSoFar);
			if (in->shouldBeExported)  addExport(newMouse);
			newLitter->addMouse(newMouse);
		}
		
		// make the rest of the mice, with the sex of each mouse randomly assigned ('u' for "unspecified")
		for (int i = 0; i < in->numRandomSex; i++, miceMadeSoFar++) {      
			Mouse *newMouse = MakeOneMouse('u', in->mother, in->father, -1, miceMadeSoFar);
			if (in->shouldBeExported) addExport(newMouse);
			newLitter->addMouse(newMouse);
		}
		
		return newLitter; // a pointer
	} catch (const string *e) {
		throw e;
	}
}


void GenomeMixer::addExport(Mouse* mousePtr) {
  // Adds this mousePtr to the list of mice to write out (to "export") to the output file.
  // Only the mice referenced in this vector will be printed out to QGene format.
  miceToExport.push_back(mousePtr); // <-- miceToExport is a vector
}



// ******** ******** ******** ******** ******** ******** ******** ******** ******** ********
//				Chromosome combination
// ******** ******** ******** ******** ******** ******** ******** ******** ******** ********
void GenomeMixer::ChromosomeCombine(Mouse *newMousePtr) {
  
  if (this->recombModel == kNO_RECOMB) {
    // If there is no recombination, then we just "recombine"
    // with the dummy function below.
    ChromosomeCopyNoRecomb(newMousePtr);
    return;
  }
  
  MarkerCollection *collectionPtr = this->markerColPtr;
  Mouse *mother = newMousePtr->motherPtr();
  Mouse *father = newMousePtr->fatherPtr();
  
  //# basic error checking: mice must be of correct gender
  if ( !(mother->isFemale() && father->isMale()) ) {
    cerr << "Error! Source mice are incorrect sexes (either the first mouse specified--the mother--is not female, or the second mouse--the father--is not male." << endl;
    exit(1);
  }
  for (int chr = 0; chr < Chromosome::xChrIndex(); chr++) {
    newMousePtr->setChr(kFEMALE, chr, combineParentalChr(chr, mother, collectionPtr));
    newMousePtr->setChr(kMALE  , chr, combineParentalChr(chr, father, collectionPtr));
  }
  if (newMousePtr->isFemale()) {
    // NewMousePtr points to a female mouse... (get X from mother, X from father)	
    newMousePtr->setChr(kFEMALE, Chromosome::xChrIndex(), combineParentalChr(Chromosome::xChrIndex(), mother, collectionPtr)); // <-- X (from mother) <-- Note that recombination happens in the X chromosome here.
    copyChr(Chromosome::xChrIndex(), newMousePtr, father, kFEMALE, kMALE); // <-- X (from father)
  } else {
    // NewMousePtr points to a male mouse... (get X from mother, Y from father)
    copyChr(Chromosome::xChrIndex(), newMousePtr, mother, kFEMALE, kFEMALE); // <-- X (from mother)
    copyChr(Chromosome::yChrIndex(), newMousePtr, father, kMALE  , kMALE); // <-- Y (from father)
  }
}



void GenomeMixer::ChromosomeCopyNoRecomb(Mouse *newMousePtr) {
	// This function is like ChromosomeCombine, except that instead of
	// doing recombination, it just straight-copies one chromosome from the mother and one from the father
	Mouse *mother = newMousePtr->motherPtr();
	Mouse *father = newMousePtr->fatherPtr();
	if ( !(mother->isFemale() && father->isMale()) ) {
		cerr << "Error! Source mice are incorrect sexes (either the first mouse specified--the mother--is not female, or the second mouse--the father--is not male." << endl;
		exit(1);
	}
	for (int chr = 0; chr < Chromosome::xChrIndex(); chr++) {
	copyChr(chr, newMousePtr, mother, kFEMALE, kFEMALE); // <-- this is a straight copy, for debugging
	copyChr(chr, newMousePtr, father, kMALE, kMALE); // <-- this is a straight copy, for debugging
	}
	if (newMousePtr->isFemale()) {
	copyChr(Chromosome::xChrIndex(), newMousePtr, mother, kFEMALE, kFEMALE); // <-- this is a straight copy, for debugging
	copyChr(Chromosome::xChrIndex(), newMousePtr, father, kFEMALE, kMALE); // <-- X (from father)
	} else {
	copyChr(Chromosome::xChrIndex(), newMousePtr, mother, kFEMALE, kFEMALE); // <-- X (from mother)
	copyChr(Chromosome::yChrIndex(), newMousePtr, father, kMALE  , kMALE); // <-- Y (from father)
	}
}


void GenomeMixer::copyChr(int chrNum, Mouse *destMouseRef, Mouse *sourceMouseRef, bool sourceSide, bool destSide) {
	// Copies a chromosome from sourceMouseRef into newMouseRef
	// The chromosome to copy is specified by chrNum and sourceSide. It is copied into chrNum and destSide.
	// sourceSide and destSide can be kMALE or kFEMALE (boolean constants).
	for (unsigned int i = 0; i < sourceMouseRef->getChr(sourceSide, chrNum)->numGeneRanges(); i++) {
		destMouseRef->getChr(destSide, chrNum)->addCopyOfGR(sourceMouseRef->getChr(sourceSide, chrNum)->getGRfromIndex(i));
	} // no return value--modifies the mouse that newMouseRef points to--that's why you pass in a reference
}



Chromosome* GenomeMixer::combineParentalChr(int chrNum, Mouse *parentPtr, MarkerCollection *markerColPtr) {
	bool side;
	Chromosome *thisChrPtr, *otherChrPtr;
	
	vector<long> *recombLocsPtr = pickRecombLocs(chrNum, markerColPtr); // returns a SORTED vector of recombination locations
	// If the recombLocs vector is not already sorted, things will be VERY BAD. It should have been sorted already, though.
	
	// Break up the two parental chromosomes at the @recombLocs
	Chromosome *matChrSegmented = segmentChr(parentPtr->getChr(kFEMALE, chrNum), recombLocsPtr);
	Chromosome *patChrSegmented = segmentChr(parentPtr->getChr(kMALE  , chrNum), recombLocsPtr);
	Chromosome *combinedChr = new Chromosome(chrNum);
	
	if (rand() < RAND_MAX/2)	side = kMALE;    // "Randomly pick either the maternal or paternal
	else			side = kFEMALE;  //  chromosome to start from (before recombination)."
	
	if (Mouse::isFemale(side)) { // getting info from maternal chromosome
		thisChrPtr  = matChrSegmented;
		otherChrPtr = patChrSegmented;
	} else { // getting info from paternal chromosome
		thisChrPtr  = patChrSegmented;
		otherChrPtr = matChrSegmented;
	}
	
	unsigned long recombIndex = 0, k = 0;
	
	while ( k < thisChrPtr->numGeneRanges() ) {
		combinedChr->addCopyOfGR(thisChrPtr->getGRfromIndex(k)); // Add
		
		if ((recombIndex < recombLocsPtr->size()) && 
			(thisChrPtr->getGRfromIndex(k)->end() + 1 == (*recombLocsPtr)[recombIndex]) ) {
			// This part Deals With Recombination.
			// We doen't handle a recombination at the very beginning of the chromosome, because that's the same as just starting from the other chromosome.
			k = otherChrPtr->getIndexFromActual((*recombLocsPtr)[recombIndex]); // Fix the new $k--because we are switching chromosomes, and the other one might have different numbers of geneRanges, so we need to make sure that switching chromosomes doesn't jump forward or backward by a bunch of bases.
			side = !side;  // Switch sides; the next item will come from the other chromosome...
			recombIndex++; // Ok, we dealt with it.
			
		} else {
			k++; // No recombination, so get the next element.
		}
		
		if (Mouse::isFemale(side)) {
			// getting info from maternal chromosome
			thisChrPtr  = matChrSegmented;
			otherChrPtr = patChrSegmented;
		} else {
			// getting info from paternal chromosome
			thisChrPtr  = patChrSegmented;
			otherChrPtr = matChrSegmented;
		}
		
	}
	
	return combinedChr; // a pointer, actually
}    


Chromosome* GenomeMixer::segmentChr(Chromosome *sourceChr, vector<long> *breaksArrayPtr) { // (called by combineParentalChr)
	// SegmentChr returns a copy of $sourceChr that has been broken at the points specified in @breaksArray
	// Example: SourceChr is:
	//		<A : 0 to 500>, <B : 501 to 1000>
	// BreaksArray is [25, 501, 1000]
	// We add breaks at 25, 501, and 1000, and the returned chromosome is then:
	//		<A : 0 to 24>, <A : 25 to 500>, <B : 501 to 999>, <B : 1000 to 1000>
	
	// So if we are given this:
	// [AAAAAAAAAAA][BBBBBBBBBBBBBBBBBBBBBB][CCCCCCCC]
	//        |            |         |
	//      break        break     break (breaks listed in @breaksArray)
	// We return this:     |         |
	//        |            |         |
	// [AAAAA][AAAA][BBBBB][BBBBBBBB][BBBBB][CCCCCCCC]
	
	// @breaksArray must be an array of integers (breakpoints, given the gene "actual" location)
	// and it must be **SORTED** in ASCENDING ORDER. Otherwise, this function will break.
	Chromosome *segmentedChr = new Chromosome(sourceChr->num());
	unsigned int breakIndex = 0;
	
	for (unsigned int i = 0; i < sourceChr->numGeneRanges(); i++) {
		// Go through all of the array
		GeneRange *theGeneRange = sourceChr->getGRfromIndex(i); // we know this exists, or we would have exited the loop
		
		if (breakIndex < breaksArrayPtr->size() && theGeneRange->contains((*breaksArrayPtr)[breakIndex]) ) {
			// If we still have breaks to process, AND this geneRange contains the breakpoint	
			// ---][AAAAAAAAAAAAA][---------][-----][--]
			//     ^      ^      ^   <-- possible locations for a break if the range "contains" a break
			//     |      |      |
			for (unsigned int k = breakIndex; k < breaksArrayPtr->size(); k++) {
				
				assert(k < breaksArrayPtr->size());
				
				if ( theGeneRange->contains((*breaksArrayPtr)[k]) ) {
					// =======================
					// This range contains the break
					if ( (*breaksArrayPtr)[k] == theGeneRange->start() ) {
						// If the break starts at the same place as an existing break, then don't do anything
					} else {
						// Otherwise, this break needs to be taken care of.
						// <   new  range   >
						// [AAAAAAAAAAAAAAAA]*................]
						// |                 |                |
						// old start     breakpoint        old ending
						// ->start()     $breaksArray[$k]    ->end()
						segmentedChr->addGR(theGeneRange->start(), 
											(*breaksArrayPtr)[k] - 1, // <-- endpoint is right before the break
											theGeneRange->value());
						// Update the GeneRange so that it's now only the part after the break.
						// <----- Old GeneRange -------->
						//               <-- new range ->
						// [ZZZZZZZZZZZ][zzzzzzzzzzzzzzz]
						//             | \  .
						//             |  \  .
						//             |   $breaksArray[$k] (start of new range)
						//             |
						// (breaksArray[k] - 1) (end of old range)
						theGeneRange = new GeneRange((*breaksArrayPtr)[k], theGeneRange->end(), theGeneRange->value());
					}
					
					if (k+1 >= breaksArrayPtr->size()) {
						// If this is the last break for this chromosome.
						segmentedChr->addGR((*breaksArrayPtr)[k], theGeneRange->end(), theGeneRange->value());
						
					} else {
						// This is not the last break, there's another one after it
						assert( (k+1) < breaksArrayPtr->size() );
						if (theGeneRange->end() < (*breaksArrayPtr)[k+1]) {
							// If this GeneRange ends before the next break...
							// [ZZZZZ][zzzzzThis_GeneRangezzzzzzz][yyyyyyy]
							//                                        L__next break
							segmentedChr->addGR((*breaksArrayPtr)[k], theGeneRange->end(), theGeneRange->value());
						} else {
							// If the next break is within this GeneRange...
							// [ZZZZZ][zzzzzThis_GeneRangezzzzzzz][yyyyyyy]
							//                  L__next break
							if ((*breaksArrayPtr)[k] == (*breaksArrayPtr)[k+1]) {
								// same breakpoint, so don't add a GR
							} else {										
								segmentedChr->addGR((*breaksArrayPtr)[k], (*breaksArrayPtr)[k+1] - 1, theGeneRange->value());
							}
							theGeneRange = new GeneRange((*breaksArrayPtr)[k+1], theGeneRange->end(), theGeneRange->value());
						}
					}
					breakIndex++; // <-- only if we contain the break do we change this...
				} // <-- end if
			} // <-- end "for"
			
		} else {
			// else, (there are no breaks in this GeneRange, just copy this GeneRange right over)
			segmentedChr->addCopyOfGR(theGeneRange); // This range doesn't contain a breakpoint, so we can add it without doing anything special
		}
	}
	
	//segmentedChr->printGeneRanges();
	//GenomeMixer::printOutRecombLocs(breaksArrayPtr, sourceChr->num());
	if (!(segmentedChr->getGRfromIndex(segmentedChr->numGeneRanges()-1)->end() == segmentedChr->lengthBases())) {
		cout << "ERROR: The last geneRange (up to " << segmentedChr->getGRfromIndex(segmentedChr->numGeneRanges()-1)->end()
			<< ") does not reach the end of the chromosome (" << segmentedChr->lengthBases() << ")" << endl;
		assert(1 == 0);
	}    
	return segmentedChr;
} // end of segmentChr



// ******** ******** ******** ******** ******** ******** ******** ******** ******** ******** ********
//			Pick Recombination Locations
// ******** ******** ******** ******** ******** ******** ******** ******** ******** ******** ********

// Given the number of the chromosome, picks out points of recombination.
// Returns a sorted array, the list of points.

// Here is where we generate the points where recombination is to occur.
// The actual recombination is handled in the calling function, so here
// all we need to do is decide where we want the recombinations to occur and put
// those locations into the "recombLocs" array.

// What we want:
//	1. "mandatory" recombination per tetrad
//	2. each centimorgan = 1 % chance of recombination
//	3. use chromosome length in cM to determine number of recombinations
//	4. on average, a 100 cM length chromosome has one recombination
//	5. in general 2 megabases ~= 1 cM in mice. This is an average.

// Map distance from the Carter-Falconer formula (below):
// my($r) = 0.25;
// my($m_fc) = (atan2((2 * $r), 1) + (log(1 + 2*$r) - log(1 - 2*$r)) / 2 ) / 4;	// more comments --> FC is for the Carter-Falconer mapping function (www.informatics.jax.org/silver/7.2.shtml) Equation: m(fc) = 1/4 { 1/2 [ln( 1+2r) - ln(1 - 2r)] + arctan(2r) }		<-- not currently being used in any form
// >>> Hey! Perl's "log" function is "ln" (log base 'e'), it's NOT log base 10. <<<
// Carter, T. C., and Falconer, D. S. (1951). Stocks for detecting linkage in the mouse and the theory of their design. J. Genet. 50: 307-323.

// IMPORTANT:
// There can only be one recombination between markers in the simulation, so markers must be fairly closely spaced.
// If you have a distance of, say, 80 cM between markers (with no intermediate markers),
// there could be two recombinations in that range in reality, but not in our simulation.

// Additionally, there should be a marker at the beginning and end of the chromosome, because recombination only happens *between* markers in this simulation.

vector<long>* GenomeMixer::pickRecombLocs(int thisChrNum, MarkerCollection *markerColPtr) {
	vector<long> *recombLocsPtr = new vector<long>; // an array of the recombination locations, in "actual" gene values, not indices. Values could be 1004003 or 9991930, for example. Not 0, 1, 2, 3... etc.	
	
	long markersThisChr = markerColPtr->markersInChr(thisChrNum);    
	for (long i = 1; i < markersThisChr; i++) {
		// int i = 1; <-- Start i at one (NOT zero), because we don't want to deal with the first marker, as it is supposed to be the very beginning of the chromosome
		long sizeOfRange = ( markerColPtr->base(thisChrNum, i) - markerColPtr->base(thisChrNum, i-1) );
		
		// There is a marker before this marker, and a marker after this one as well.
		if (sizeOfRange > 0) {
			// As long as the i'th marker comes after the (i-1)'th in cM measurement, do this...
			
			if ( hasARecomb(markerColPtr->cM(thisChrNum, i) - markerColPtr->cM(thisChrNum, i - 1)) ) {
				if ((markerColPtr->base(thisChrNum, i) - markerColPtr->base(thisChrNum, i-1)) <= 0) {
					cerr << ">>> Error, two items have the same location (in bases) on the same chromosome." << endl;
					exit(1);
				}
				
				long randomToSize = long((rand()/(RAND_MAX + 1.0)) * sizeOfRange); // <-- gives a random integer between 0 and sizeOfRange
				long proposedBreak = randomToSize + markerColPtr->base(thisChrNum, i-1);
				// put a proposed breakpoint somewhere in that interval
				// We say "proposed" because now we're going to check for interference, and the recombination might be disqualified.
				if (checkProposedPoint(proposedBreak, recombLocsPtr)) {
					// If we are here, then the proposed point was far enough away from any other breakpoints that we don't worry about interference.
					recombLocsPtr->push_back(proposedBreak);
				}
			}
			
		} else {
			cout << ">>> Warning: two markers have the same cM position on chromosome " << Chromosome::chrNameFromIndex(thisChrNum) << "." << endl;
		}	    
	} // end of "for" loop
	
	std::sort(recombLocsPtr->begin(), recombLocsPtr->end()); // <-- this sorts the array (of long ints) in ascending numerical order    
	//GenomeMixer::printOutRecombLocs(recombLocsPtr, thisChrNum); // <-- for debugging
	return recombLocsPtr; // Warning: The return value MUST be sorted before it is returned.
}



void GenomeMixer::printOutRecombLocs(vector<long> *recombLocsPtr, int thisChrNum) {
	cout << "@@ Printing out recombination locations (in number of bases)." << endl;
	for (unsigned int i = 0; i < recombLocsPtr->size(); i++) {
		cout << "@@  Breakpoint at " << (*recombLocsPtr)[i] << " on " << Chromosome::chrNameFromIndex(thisChrNum) << endl;
	}
	cout << "@@ End of recombination locations." << endl << endl;
}



bool GenomeMixer::checkProposedPoint(long thisPoint, vector<long> *vecOfBreakPts) {
	// Interference detector: checks a proposed breakpoint.
	
	// Right now, we have absolute interference over $noRecombDistBases on either direction,
	// and then no interference. There is no gradual change as we get farther away from a breakpoint.
	// But there could be, if this function was changed to accomodate a more sophisticated interference function.
	
	// 1 cM ~= 2 MB in mouse		(1 cM ~= 2000 KBases (2 MBases): different from chr to chr, male/female, however)
	// 10 cM ~= 20 MB in mouse
	
	// This function just claims that there can be no recombination within $noRecombDistBases bases from another recombination.
	// The initial value was 25000000 (25,000,000 bases on either side of a recombination, in which no other recombination may occcur)
	
	for (unsigned int i = 0; i < vecOfBreakPts->size(); i++) {
		if ( thisPoint != (*vecOfBreakPts)[i] && 		// <-- don't check this point against itself
			 thisPoint+noRecombDistBases > (*vecOfBreakPts)[i] &&
			 thisPoint-noRecombDistBases < (*vecOfBreakPts)[i] ) {
			return false;
		}
	}
	return true;
}


bool GenomeMixer::hasARecomb(float cM_dist) {
  double probOfRecomb;

  // This doesn't take interference into account, but we do that ourselves. Even if
  // "has a recomb" says "yes, we have a recombination here," we will still go and check around that recombination
  // to see if there is already another recombination that intereferes with it. If so, we then don't put in this new recombination.
  
  switch (this->recombModel) {
  case kNO_RECOMB:
    return false; // no recombinations ever...
    break;
  case kLIN_RECOMB:
    return (cM_dist > (rand()/(RAND_MAX + 1.0) * 100.0));
    // 1% for 1cM, 2% for 2cM, 100% for 100cM and greater...
    break;
  case kHALDANE_RECOMB:
    probOfRecomb = 0.5 * (1 - std::exp(cM_dist / 100.0 * -2));
    return (probOfRecomb > ((double) rand() / ((double) RAND_MAX)));
    // std::exp(something)  is "e to the (something) power" (returning a double)      
    // since a centiMorgan = 1/100 of a morgan, we divide the cM_dist by 100.0 here.
    // the formula wants the distance in Morgans, remember.
    // Haldane Formula is (prob of recomb) = 0.5 * (1 - e^(-2 * morgan_distance))
    // Here are some sample recombination frequencies with this formula:
    // cM_dist 	morgan_dist 	resulting prob of recomb in interval
    //    0 		0 			0 (as it should be)
    //    1 		0.01			0.0099 (almost 1%)
    //    10 		0.10 		0.09 (about 9%)
    //    50 		0.5 			0.316 
    // And the probability asymptotically approaches 0.5 as the cm_dist increases
    
    break;
  default:
    throw new string("Error in recombination: invalid model selected!");
    break;
  }
  
}

